//------------------------------------------------------------------------
//
// Copyright (C) 1997-2009 iNTERFACEWARE Inc.  All Rights Reserved
//
// Module:  C++ Source Plugin Example
// 
// Description
//
// This application is a simple example of using the Chameleon generated 
// objects within an Iguana plugin.  It is intended to be a helpful starting 
// point in understanding how to use these objects but must be fleshed out 
// in the following areas to provide a full solution for an application 
// HL7 interface:
//
// 1) The code will need to be modified to retrieve message data from a 
//    database whether by using explicit database calls or going through 
//    some type of middle tier layer that your application may support.
//
// 2) Definitely take the time to read the manual and go through the 
//    tutorials after looking at this example before attempting to deploy 
//    a production solution.  We do provide custom on-site training and 
//    technical implementation services if you require them.
//
// Author: iNTERFACEWARE
// Date:   Wed 08/17/2009
//---------------------------------------------------------------------------
#include "CHMclientPreCompHead.h"
#pragma hdrstop

#include <IGC/IGCdisconnectReason.h>
#include <IGC/IGCsource.h>
#include <IGC/IGCsourceMessageRequestedEventInputs.h>
#include <IGC/IGCsourceMessageRequestedEventOutputs.h>
#include <IGC/IGCsourceMessageRequestCompletedEventInputs.h>
#include <IGC/IGCsourceMessageRequestCompletedEventOutputs.h>
#include <IGC/IGCerrorClass.h>

#include <iostream>
using namespace std;

/* template code */

class FromPluginExample: public IGCsource
{
public:
   // Simple logging class: outputs both to stdout and a string.
   class OutputLog : public CHMoutputLog
   {
   public:
      virtual void OutputLine(const CHMchar* Line)
      {
         m_LogString.append(Line);
         m_LogString.append("\n");

         cout << Line << endl;
      }

      CHMstring m_LogString;
   };
   
   FromPluginExample(): IGCsource()
   {
      cout << "Trying to load test_data.vmd." << endl;

      m_Engine.Load("test_data.vmd");
      m_Engine.SetLog(m_pEngineOutputLog = new OutputLog);
      m_Engine.SetCurrentConfigurationIndex(0);

      cout << "Loaded message definition file test_data.vmd." << endl;
   }

   ~FromPluginExample()
   {
   }

   PrefixTestappEngine m_Engine;
   OutputLog* m_pEngineOutputLog; // Owned (deleted) by m_Engine
   CHMdateTime m_LastMessageRequestTime;

   CHMstring GenerateRandomMessage()
   {
      // Begin message generation code (comment used as an anchor for STG) 
      switch (rand() % 1)
      {
      case 0:
         return m_Engine.GenerateADT_A04();
      default:
         return m_Engine.GenerateADT_A04();
      }
      // End message generation code (comment used as an anchor for STG) 
   }

   // This event is triggered when Iguana requests a message.
   virtual void OnMessageRequested
   (
      IGCsourceMessageRequestedEventInputs&, // Unused
      IGCsourceMessageRequestedEventOutputs& Outputs
   )
   {
      // Note that exceptions can never be thrown outside the handler.
      try
      {
         try
         {
            cout << endl;
            cout << "*** Message Requested ***" << endl;

            // To simulate the behavior of a real-world plugin, we throttle the rate
            // at which we generate messages to at most once per second.
            //
            // Real world plugins need no such throttling mechanism.  Real-world plugins
            // should respond to message requests immediately with either the next available 
            // message, or an empty response if no messages are presently available.
            //
            // When a plugin responds to a message request with an empty response (signifying
            // that no messages are presently available), Iguana will delay for the period 
            // specified by the channel's From Plugin component Polling Time setting before 
            // sending another message request.

            const CHMdateTime Now = CHMdateTime::Now();
            CHMboolean GenerateMessage = Now.Second() != m_LastMessageRequestTime.Second();
            m_LastMessageRequestTime = Now;

            if (GenerateMessage)
            {
               // Set OutboundMessage to the next available message.
               m_pEngineOutputLog->m_LogString.clear();
               CHMstring OutboundMessage = GenerateRandomMessage();
               Outputs.SetOutboundMessage(OutboundMessage.c_str());

               // Set PluginMessageId to a unique identified for the message, which will
               // be sent back to the plugin in the next OnMessageRequestCompleted event.
               CHMstring PluginMessageId = m_LastMessageRequestTime.Dump();
               Outputs.SetPluginMessageId(PluginMessageId.c_str());

               cout << "Message Generated (plugin message id: " << PluginMessageId.c_str() << ")" << endl;
               cout << CHMlineFeedConvert(OutboundMessage.c_str()).c_str() << endl;
            }
            else
            {
               cout << "Responding with an empty message - no messages presently available." << endl;
            }
         }
         catch (const CHMerror& Error)
         {
            cout << "Message Generation Error: " << Error.Description() << endl;

            // Add an error-level log message to Iguana's logs describing the failure.
            Outputs.AddErrorLogMessage(Error.Description());

            // You could also use StopChannel, StopChannelErrorFlag, and StopChannelReason 
            // to stop the Iguana channel with or without an error flag, and provide a 
            // reason for the stop request.
            //
            //    Response.SetStopChannel(true);
            //    Response.SetStopChannelErrorFlag(true);
            //    Response.SetStopChannelReason("An error occurred.");
         }

         // You can send additional error/warning/info/debug log messages to Iguana.  In this case
         // we simply send an info-level log message containing the Chameleon engine's log output.
         if (!m_pEngineOutputLog->m_LogString.is_null())
         {
            Outputs.AddInfoLogMessage(m_pEngineOutputLog->m_LogString.c_str());
         }
      }
      catch (const IGCerror& Error)
      {
         cout << "Error: " << Error.Description() << endl;
      }
      catch (...)
      {
         cout << "Error: unexpected exception" << endl;
      }
   }

   // This event is triggered when Iguana has received the message sent via OnMessageRequest.
   virtual void OnMessageRequestCompleted
   (
      IGCsourceMessageRequestCompletedEventInputs& Inputs,
      IGCsourceMessageRequestCompletedEventOutputs& // Unused
   )
   {
      // Note that exceptions can never be thrown outside the handler.
      try
      {
         // PluginMessageId will contain the unique identifier provided by the plugin in the
         // previous call to OnMessageRequest.
         //
         // If MessageProcessed is false, Iguana was unable to enqueue the message, and it
         // should be sent again.

         cout << endl;
         cout << "*** Message Request Completed (plugin message id: " << Inputs.PluginMessageId() << ") ***" << endl;
      }
      catch (...)
      {
         cout << "Error: unexpected exception" << endl;
      }
   }

   // This event is triggered when connection to Iguana is closed.
   // See also the use of WaitForDisconnect() below.
   virtual void OnDisconnected()
   {
      // Note that exceptions can never be thrown outside the handler.
      try
      {
         cout << "Source disconnected." << endl;
      }
      catch (...)
      {
         cout << "Error: unexpected exception" << endl;
      }
   }
};

int main(int argc, const char** argv)
{
   try
   {
      // Disable buffering for output streams, so that Iguana receives
      // output immediately rather than when the output buffer happens
      // to be filled.  Note that this function must be called before
      // any output is written.  An alternative to this approach would
      // be to call fflush() judiciously after all output commands.
      IGC_DISABLE_OUTPUT_BUFFERING();

      if (argc < 4)
      {
         throw IGCerror("Invalid command line arguments.", __LINE__, __FILE__);
      }

      const char* pHostname    = argv[1];
      int         Port         = atoi( argv[2] );
      const char* pChannel     = argv[3];

      // Connect to Iguana.
      FromPluginExample Source;
      cout << "Connecting to Iguana" << endl;
      Source.Connect(pHostname, Port, pChannel);

      // Wait for the plugin to disconnect.
      // Could also use WaitForDisconnectWithTimeout() here.
      Source.WaitForDisconnect();

      // Print the reason for disconnection.
      IGCdisconnectReason DisconnectReason = Source.DisconnectReason();
      cout << DisconnectReason.Description() << endl;

      return 0;
   }
   catch (const IGCerror& Error)
   {
      cout << "Error: " << Error.Description() << endl << endl;
   }
   catch (const CHMerror& Error)
   {
      cout << "Error: " << Error.Description() << endl << endl;
   }
   catch (...)
   {
      cout << "Error: unexpected exception" << endl << endl;
   }

   cout << "This is an Iguana source plugin.  It runs as a command line application\n"
           "which connects via TCP/IP to the plugin communication server within Iguana.\n"
           "It will be necessary to run a channel in Iguana and set up a channel with\n"
           "a From Plugin component for this to connect to.  See:\n"
           "\n"
           "   http://www.interfaceware.com/manual/plugin_setup_cpp.html\n"
           "\n"
           "Once the channel is set up and running you can connect to it with this usage:\n"
           "\n"
           "   " << argv[0] << " <iguana hostname> <plugin communication port> <channel name>\n"
           "\n"
           "The plugin communication port is the port that Iguana listens on for plugin\n"
           "communication." << endl;

   return EXIT_FAILURE;
}
